@@ -1,5 +1,5 @@
 /* GAIL - The GNOME Accessibility Enabling Library
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -89,8 +89,6 @@ gail_scrolled_window_new (GtkWidget *widget)
   accessible = ATK_OBJECT (object);
   atk_object_initialize (accessible, widget);
 
-  accessible->role = ATK_ROLE_SCROLL_PANE;
-
   return accessible;
 }
 
@@ -109,6 +107,8 @@ gail_scrolled_window_real_initialize (AtkObject *obj,
   g_signal_connect_data (window->vscrollbar, "notify::visible",
     (GCallback)gail_scrolled_window_scrollbar_visibility_changed, 
     obj, FALSE, FALSE);
+
+  obj->role = ATK_ROLE_SCROLL_PANE;
 }
 
 static gint
@@ -128,7 +128,7 @@ gail_scrolled_window_get_n_children (AtkObject *object)
    
   /* Get the number of children returned by the backing GtkScrolledWindow */
 
-  children = gtk_container_children (GTK_CONTAINER(gtk_window));
+  children = gtk_container_get_children (GTK_CONTAINER(gtk_window));
   n_children = g_list_length (children);
   g_list_free (children);
   
@@ -160,7 +160,7 @@ gail_scrolled_window_ref_child (AtkObject *obj,
 
   gtk_window = GTK_SCROLLED_WINDOW (widget);
 
-  children = gtk_container_children (GTK_CONTAINER (gtk_window));
+  children = gtk_container_get_children (GTK_CONTAINER (gtk_window));
   n_children = g_list_length (children);
 
   if (child == n_children)
@@ -170,16 +170,21 @@ gail_scrolled_window_ref_child (AtkObject *obj,
       else if (gtk_window->vscrollbar_visible)
         accessible = gtk_widget_get_accessible (gtk_window->vscrollbar);
     }
-  else if (child == n_children+1 && gtk_window->vscrollbar_visible)
+  else if (child == n_children+1 && 
+           gtk_window->hscrollbar_visible &&
+           gtk_window->vscrollbar_visible)
     accessible = gtk_widget_get_accessible (gtk_window->vscrollbar);
   else if (child < n_children)
     {
       tmp_list = g_list_nth (children, child);
-      accessible = gtk_widget_get_accessible (GTK_WIDGET (tmp_list->data));
+      if (tmp_list)
+	accessible = gtk_widget_get_accessible (
+		GTK_WIDGET (tmp_list->data));
     }
 
   g_list_free (children);
-  g_object_ref (accessible);
+  if (accessible)
+    g_object_ref (accessible);
   return accessible; 
 }
 
@@ -200,8 +205,9 @@ gail_scrolled_window_scrollbar_visibility_changed (GObject    *object,
       gchar *signal_name;
 
       gtk_window = GTK_SCROLLED_WINDOW (GTK_ACCESSIBLE (user_data)->widget);
-      children = gtk_container_children (GTK_CONTAINER (gtk_window));
+      children = gtk_container_get_children (GTK_CONTAINER (gtk_window));
       index = n_children = g_list_length (children);
+      g_list_free (children);
 
       if ((gpointer) object == (gpointer) (gtk_window->hscrollbar))
         {
