@@ -258,7 +258,7 @@ gail_window_real_initialize (AtkObject *obj,
     obj->role = ATK_ROLE_ALERT;
   else if (GTK_IS_DIALOG (widget))
     obj->role = ATK_ROLE_DIALOG;
-  else if (GTK_IS_WINDOW (widget))
+  else
     {
       const gchar *name;
 
@@ -267,13 +267,11 @@ gail_window_real_initialize (AtkObject *obj,
         obj->role = ATK_ROLE_TOOL_TIP;
       else if (GTK_IS_PLUG (widget))
         obj->role = ATK_ROLE_PANEL;
+      else if (GTK_WINDOW (widget)->type == GTK_WINDOW_POPUP)
+        obj->role = ATK_ROLE_WINDOW;
       else
         obj->role = ATK_ROLE_FRAME;
     }
-  else if (GTK_IS_HANDLE_BOX (widget))
-    obj->role = ATK_ROLE_UNKNOWN;
-  else
-    obj->role = ATK_ROLE_INVALID;
 }
 
 static void
